Package: isoband
Title: Generate Isolines and Isobands from Regularly Spaced Elevation
        Grids
Version: 0.2.7
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-4757-117X")),
    person("Claus O.", "Wilke", , "wilke@austin.utexas.edu", role = "aut",
           comment = c("Original author", ORCID = "0000-0002-7470-9261")),
    person("Thomas Lin", "Pedersen", , "thomasp85@gmail.com", role = "aut",
           comment = c(ORCID = "0000-0002-5147-4711"))
  )
Description: A fast C++ implementation to generate contour lines
    (isolines) and contour polygons (isobands) from regularly spaced grids
    containing elevation data.
License: MIT + file LICENSE
URL: https://isoband.r-lib.org
BugReports: https://github.com/r-lib/isoband/issues
Imports: grid, utils
Suggests: covr, ggplot2, knitr, magick, microbenchmark, rmarkdown, sf,
        testthat, xml2
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.2.3
SystemRequirements: C++11
NeedsCompilation: yes
Packaged: 2022-12-19 20:10:02 UTC; hadleywickham
Author: Hadley Wickham [aut, cre] (<https://orcid.org/0000-0003-4757-117X>),
  Claus O. Wilke [aut] (Original author,
    <https://orcid.org/0000-0002-7470-9261>),
  Thomas Lin Pedersen [aut] (<https://orcid.org/0000-0002-5147-4711>)
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2022-12-20 10:00:13 UTC
Built: R 4.4.0; amd64-portbld-midnightbsd4; 2026-07-18 02:00:58 UTC; unix

Package: isoband
Title: Generate Isolines and Isobands from Regularly Spaced Elevation
        Grids
Version: 0.3.0
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = "aut",
           comment = c(ORCID = "0000-0003-4757-117X")),
    person("Claus O.", "Wilke", , "wilke@austin.utexas.edu", role = "aut",
           comment = c("Original author", ORCID = "0000-0002-7470-9261")),
    person("Thomas Lin", "Pedersen", , "thomas.pedersen@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-5147-4711")),
    person("Posit, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: A fast C++ implementation to generate contour lines
    (isolines) and contour polygons (isobands) from regularly spaced grids
    containing elevation data.
License: MIT + file LICENSE
URL: https://isoband.r-lib.org, https://github.com/r-lib/isoband
BugReports: https://github.com/r-lib/isoband/issues
Imports: cli, grid, rlang, utils
Suggests: covr, ggplot2, knitr, magick, bench, rmarkdown, sf, testthat
        (>= 3.0.0), xml2
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-12-05
Encoding: UTF-8
RoxygenNote: 7.3.3
Config/build/compilation-database: true
LinkingTo: cpp11
NeedsCompilation: yes
Packaged: 2025-12-05 12:52:07 UTC; thomas
Author: Hadley Wickham [aut] (ORCID: <https://orcid.org/0000-0003-4757-117X>),
  Claus O. Wilke [aut] (Original author, ORCID:
    <https://orcid.org/0000-0002-7470-9261>),
  Thomas Lin Pedersen [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-5147-4711>),
  Posit, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Thomas Lin Pedersen <thomas.pedersen@posit.co>
Repository: CRAN
Date/Publication: 2025-12-07 06:10:07 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-27 05:37:32 UTC; unix
Archs: isoband.so.dSYM

Package: isoband
Title: Generate Isolines and Isobands from Regularly Spaced Elevation
        Grids
Version: 0.2.5
Authors@R: 
    c(person(given = "Claus O.",
             family = "Wilke",
             role = c("aut", "cre"),
             email = "wilke@austin.utexas.edu",
             comment = c(ORCID = "0000-0002-7470-9261")),
      person(given = "Thomas Lin",
             family = "Pedersen",
             role = "aut",
             email = "thomasp85@gmail.com",
             comment = c(ORCID = "0000-0002-5147-4711")),
      person(given = "testthat and Catch authors",
             role = "ctb",
             comment = "testthat C++ testing code"))
Description: A fast C++ implementation to generate contour lines (isolines) and
  contour polygons (isobands) from regularly spaced grids containing elevation data.
URL: https://wilkelab.org/isoband/
BugReports: https://github.com/wilkelab/isoband/issues
License: MIT + file LICENSE
Encoding: UTF-8
Imports: grid, utils
RoxygenNote: 7.1.1
Config/testthat/edition: 3
Suggests: covr, ggplot2, knitr, magick, microbenchmark, rmarkdown, sf,
        testthat, xml2
SystemRequirements: C++11
VignetteBuilder: knitr
NeedsCompilation: yes
Packaged: 2021-07-12 20:44:13 UTC; clauswilke
Author: Claus O. Wilke [aut, cre] (<https://orcid.org/0000-0002-7470-9261>),
  Thomas Lin Pedersen [aut] (<https://orcid.org/0000-0002-5147-4711>),
  testthat and Catch authors [ctb] (testthat C++ testing code)
Maintainer: Claus O. Wilke <wilke@austin.utexas.edu>
Repository: CRAN
Date/Publication: 2021-07-13 04:50:13 UTC

# Generated by roxygen2: do not edit by hand

S3method(iso_to_sfg,default)
S3method(iso_to_sfg,isobands)
S3method(iso_to_sfg,isolines)
S3method(makeContent,isobands_grob)
S3method(makeContent,isolines_grob)
S3method(makeContext,isolines_grob)
export(angle_fixed)
export(angle_halfcircle_bottom)
export(angle_halfcircle_right)
export(angle_identity)
export(clip_lines)
export(iso_to_sfg)
export(isobands)
export(isobands_grob)
export(isolines)
export(isolines_grob)
export(label_placer_manual)
export(label_placer_middle)
export(label_placer_minmax)
export(label_placer_none)
export(label_placer_simple)
export(plot_iso)
import(grid)
importFrom(utils,modifyList)
useDynLib(isoband, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# isoband <img width="120" alt="" align="right" src="man/figures/logo.png">

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/isoband)](https://CRAN.R-project.org/package=isoband)
[![R-CMD-check](https://github.com/r-lib/isoband/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/isoband/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/isoband/graph/badge.svg)](https://app.codecov.io/gh/r-lib/isoband)
<!-- badges: end -->

Generate contour lines (isolines) and contour polygons (isobands) from
regularly spaced grids containing elevation data. Package originally
written by Claus Wilke and donated to r-lib in 2022.

## Installation

Install the latest official release from CRAN via:

``` r
install.packages("isoband")
```

Install the current development from github via:

``` r
# install.packages("pak")
pak::pak("r-lib/isoband")
```

## Examples

The two main workhorses of the package are the functions `isolines()`
and `isobands()`, respectively. They return a list of isolines/isobands
for each isolevel specified. Each isoline/isoband consists of vectors of
x and y coordinates, as well as a vector of ids specifying which sets of
coordinates should be connected. This format can be handed directly to
`grid.polyline()`/`grid.path()` for drawing. However, we can also
convert the output to spatial features and draw with ggplot2 (see
below).

``` r
library(isoband)

m <- matrix(c(0, 0, 0, 0, 0,
              0, 1, 2, 1, 0,
              0, 1, 2, 0, 0,
              0, 1, 0, 1, 0,
              0, 0, 0, 0, 0), 5, 5, byrow = TRUE)

isolines(1:ncol(m), 1:nrow(m), m, 0.5)
#> $`0.5`
#> $`0.5`$x
#>  [1] 4.00 3.50 3.00 2.50 2.00 1.50 1.50 1.50 2.00 3.00 4.00 4.50 4.00 3.75 4.00
#> [16] 4.50 4.00
#> 
#> $`0.5`$y
#>  [1] 4.50 4.00 3.75 4.00 4.50 4.00 3.00 2.00 1.50 1.25 1.50 2.00 2.50 3.00 3.50
#> [16] 4.00 4.50
#> 
#> $`0.5`$id
#>  [1] 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1
#> 
#> 
#> attr(,"class")
#> [1] "isolines" "iso"

isobands(1:ncol(m), 1:nrow(m), m, 0.5, 1.5)
#> $`0.5:1.5`
#> $`0.5:1.5`$x
#>  [1] 2.50 2.00 1.50 1.50 1.50 2.00 3.00 4.00 4.50 4.00 3.75 4.00 4.50 4.00 3.50
#> [16] 3.00 3.00 3.25 3.50 3.00 2.50 2.50
#> 
#> $`0.5:1.5`$y
#>  [1] 4.00 4.50 4.00 3.00 2.00 1.50 1.25 1.50 2.00 2.50 3.00 3.50 4.00 4.50 4.00
#> [16] 3.75 3.25 3.00 2.00 1.75 2.00 3.00
#> 
#> $`0.5:1.5`$id
#>  [1] 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2
#> 
#> 
#> attr(,"class")
#> [1] "isobands" "iso"
```

The function `plot_iso()` is a convenience function for debugging and
testing.

``` r
plot_iso(m, 0.5, 1.5)
```

<img src="man/figures/README-basic-example-plot-1.png" width="50%" />

The isolining and isobanding algorithms have no problem with larger
datasets. Let’s calculate isolines and isobands for the volcano dataset,
convert to sf, and plot with ggplot2.

``` r
library(ggplot2)
suppressWarnings(library(sf))
#> Linking to GEOS 3.13.0, GDAL 3.8.5, PROJ 9.5.1; sf_use_s2() is TRUE

m <- volcano
b <- isobands((1:ncol(m))/(ncol(m)+1), (nrow(m):1)/(nrow(m)+1), m, 10*(9:19), 10*(10:20))
l <- isolines((1:ncol(m))/(ncol(m)+1), (nrow(m):1)/(nrow(m)+1), m, 10*(10:19))

bands <- iso_to_sfg(b)
data_bands <- st_sf(
  level = 1:length(bands),
  geometry = st_sfc(bands)
)
lines <- iso_to_sfg(l)
data_lines <- st_sf(
  level = 2:(length(lines)+1),
  geometry = st_sfc(lines)
)

ggplot() +
  geom_sf(data = data_bands, aes(fill = level), color = NA, alpha = 0.7) +
  geom_sf(data = data_lines, color = "black") +
  scale_fill_viridis_c(guide = "none") +
  coord_sf(expand = FALSE)
```

<img src="man/figures/README-volcano-1.png" width="50%" />

#' Render labeled isolines
#'
#' This function generates a grid grob that represents labeled isolines.
#'
#' @param lines Isolines, as produced by the [`isolines()`] function.
#' @param gp Grid graphical parameters. Parameters applying to lines
#'   (such as `col`, `lwd`, `lty`, etc.) are recycled among the total
#'   number of lines drawn. Parameters applying only to labels (such
#'   as `fontfamily`, `fontsize`) are recycled among the specified
#'   breaks only. The two parameters `col` and `alpha` are also applied
#'   to labels, unless overridden (see `label_col` and `label_alpha`),
#'   but are matched to the corresponding lines.
#' @param breaks Character vector specifying the isolines that should be
#'   labeled. If `NULL`, labels all isolines.
#' @param labels Character vector specifying the labels for each break.
#'   If `NULL`, uses the breaks as labels. The number of labels provided
#'   must match the number of breaks provided.
#' @param margin Unit object of length 4 specifying the top, right, bottom,
#'   and left margins around each text label. The same margins are applied
#'   to all labels.
#' @param label_col Color applied to labels. Can be used to override the
#'   color provided in `gp`, in case labels and lines should have different
#'   colors.
#' @param label_alpha Alpha applied to labels. Can be used to override the
#'   alpha value provided in `gp`, in case labels and lines should have
#'   different alpha values.
#' @param label_placer Function that controls how labels are placed along
#'   the isolines. Uses [`label_placer_minmax()`] by default.
#' @param units A character string specifying the units in which to
#'   interpret the isolines coordinates. Defaults to `"npc"`.
#' @seealso
#' See [`isobands_grob()`] for drawing of isobands. See [`label_placer_minmax()`] for
#' label placement strategies.
#' @examples
#' library(grid)
#'
#' viridis_pal <- colorRampPalette(
#'   c("#440154", "#414487", "#2A788E", "#22A884", "#7AD151", "#FDE725"),
#'   space = "Lab"
#' )
#'
#' x <- (1:ncol(volcano))/(ncol(volcano)+1)
#' y <- (nrow(volcano):1)/(nrow(volcano)+1)
#' lines <- isolines(x, y, volcano, 5*(19:38))
#' bands <- isobands(x, y, volcano, 5*(18:38), 5*(19:39))
#'
#' b <- isobands_grob(
#'   bands,
#'   gp = gpar(col = NA, fill = viridis_pal(21), alpha = 0.4)
#' )
#' l <- isolines_grob(
#'   lines, breaks = 20*(5:10),
#'   gp = gpar(
#'     lwd = c(.3, 1, .3, .3)
#'   )
#' )
#'
#' grid.newpage()
#' grid.draw(b)
#' grid.draw(l)
#' @export
isolines_grob <- function(lines, gp = gpar(), breaks = NULL, labels = NULL,
                          margin = unit(c(1, 1, 1, 1), "pt"), label_col = NULL, label_alpha = NULL,
                          label_placer = label_placer_minmax(), units = "npc") {
  if (is.null(breaks)) {
    breaks <- names(lines)
  } else {
    breaks <- as.character(breaks)
  }

  if (is.null(labels)) {
    labels <- breaks
  } else if (length(labels) != length(breaks)) {
    stop("Number of labels must match the number of breaks.", call. = FALSE)
  } else {
    labels <- as.character(labels)
  }

  if (length(margin) != 4 || !is.unit(margin)) {
    stop("The `margin` parameter must be a unit object of length four.", call. = FALSE)
  }

  # first set up a data frame with all the label info
  labels_data <- data.frame(
    index = match(breaks, names(lines)), # index of labeled lines in original list of lines, for matching of graphical parameters
    break_index = 1:length(breaks), # index into original list of breaks, for matching graphical parameters
    break_id = breaks, # identifier for each break (corresponds to the name in the lines column)
    label = labels, # label for each break
    stringsAsFactors = FALSE
  )
  # then calculate the position of all labels via the `label_placer()` function
  labels_data <- label_placer(lines, labels_data)

  gTree(
    lines = lines,
    breaks = breaks,
    labels = labels,
    labels_data = labels_data,
    gp_combined = gp,
    margin = margin,
    label_col = label_col,
    label_alpha = label_alpha,
    label_placer = label_placer,
    units = units,
    cl = "isolines_grob"
  )
}

#' @export
makeContext.isolines_grob <- function(x) {
  # store current graphics parameters for later
  x$gp_cur <- get.gpar()

  # if we have no labels then nothing else needs to be done
  if (nrow(x$labels_data) == 0) {
    return(x)
  }

  # we need to set up the gp slot for text labels, so font calculations work
  gp <- modifyList(x$gp_cur, x$gp_combined)

  # map graphical parameters to duplicated labels
  # we only handle font parameters here, to guarantee proper
  # calculation of label sizes
  n <- length(x$breaks)
  cex <- rep_len(gp$cex, n)[x$labels_data$break_index]
  lineheight <- rep_len(gp$lineheight, n)[x$labels_data$break_index]
  fontfamily <- rep_len(gp$fontfamily, n)[x$labels_data$break_index]
  fontsize <- rep_len(gp$fontsize, n)[x$labels_data$break_index]

  # fontface needs special treatment, since it can be NULL if font
  # is specified
  if (is.null(gp$fontface)) {
    # we work with font
    font <- rep_len(gp$font, n)[x$labels_data$break_index]

    x$gp <- gpar(
      cex = cex, fontsize = fontsize, lineheight = lineheight,
      fontfamily = fontfamily, font = font
    )
  } else {
    # we work with fontface
    fontface <- rep_len(gp$fontface, n)[x$labels_data$break_index]

    x$gp <- gpar(
      cex = cex, fontsize = fontsize, lineheight = lineheight,
      fontfamily = fontfamily, fontface = fontface
    )
  }

  x
}

#' @export
makeContent.isolines_grob <- function(x) {
  labels_data <- x$labels_data

  if (nrow(labels_data) == 0) {
    # no labels to be drawn, nothing to be done
    return(isolines_grob_makeContent_nolabels(x))
  }

  # calculate label widths and heights in npc units
  label_widths <- convertWidth(stringWidth(labels_data$label), x$units, valueOnly = TRUE)
  label_heights <- convertHeight(
    stringHeight(labels_data$label) + stringDescent(labels_data$label),
    x$units, valueOnly = TRUE
  )

  # get viewport aspect ratio to correct clipping for rotated labels
  asp <- convertHeight(unit(1, "pt"), x$units, valueOnly = TRUE) / convertWidth(unit(1, "pt"), x$units, valueOnly = TRUE)
  #print(asp)

  # calculate margins in npc units
  margin_rl <- convertWidth(x$margin[c(2, 4)], x$units, valueOnly = TRUE)
  margin_tb <- convertHeight(x$margin[c(1, 3)], x$units, valueOnly = TRUE)
  margin_w <- sum(margin_rl)
  margin_h <- sum(margin_tb)
  margin_wdiff <- margin_rl[2] - margin_rl[1]
  margin_hdiff <- margin_rl[1] - margin_rl[2]

  # calculate the clip box for each label
  # xoff and yoff are needed to correct for uneven label margins
  xoff <- -margin_wdiff*cos(labels_data$theta)/2 + (margin_hdiff/asp)*sin(labels_data$theta)/2
  yoff <- -margin_wdiff*sin(labels_data$theta)/2 - margin_hdiff*cos(labels_data$theta)/2

  clip_boxes <- data.frame(
    x = labels_data$x + xoff, y = labels_data$y + yoff,
    width = label_widths + margin_w, height = label_heights + margin_h,
    theta = labels_data$theta
  )

  make_lines_grobs <- function(data, col, alpha, lty, lwd, lex, lineend, linejoin, linemitre) {
    if (length(data$x) == 0) {
      return(NULL)
    }
    clipped <- clip_lines(data$x, data$y, data$id, clip_boxes, asp = asp)
    if (length(clipped$x) == 0) {
      return(NULL)
    }
    polylineGrob(
      clipped$x, clipped$y, clipped$id,
      default.units = x$units,
      gp = gpar(
        col = col, alpha = alpha, lty = lty, lwd = lwd, lex = lex,
        lineend = lineend, linejoin = linejoin, linemitre = linemitre
      )
    )
  }

  # merge current and grob-specific graphical parameters so we can redistribute among isolevels
  gp <- modifyList(x$gp_cur, x$gp_combined)
  n <- length(x$lines)

  lines_grobs <- mapply(
    make_lines_grobs,
    x$lines,
    rep_len(gp$col, n),
    rep_len(gp$alpha, n),
    rep_len(gp$lty, n),
    rep_len(gp$lwd, n),
    rep_len(gp$lex, n),
    rep_len(gp$lineend, n),
    rep_len(gp$linejoin, n),
    rep_len(gp$linemitre, n),
    SIMPLIFY = FALSE
  )

  # calculate color and alpha for text labels
  if (is.null(x$label_col)) {
    col <- rep_len(gp$col, length(x$lines))[x$labels_data$index]
  } else {
    col <- rep_len(x$label_col, length(x$breaks))[x$labels_data$break_index]
  }
  if (is.null(x$label_alpha)) {
    # alpha is cumulative, so we don't take it from the current viewport
    alpha <- rep_len(x$gp_combined$alpha %||% 1, length(x$lines))[x$labels_data$index]
  } else {
    alpha <- rep_len(x$label_alpha, n)[x$labels_data$break_index]
  }

  labels_grob <- textGrob(
    labels_data$label, labels_data$x, labels_data$y, rot = 360*labels_data$theta/(2*pi),
    default.units = x$units,
    gp = gpar(col = col, alpha = alpha)
  )

  children <- do.call(gList, c(lines_grobs, list(labels_grob)))
  setChildren(x, children)
}

isolines_grob_makeContent_nolabels <- function(x) {
  make_lines_grobs <- function(data, col, alpha, lty, lwd, lex, lineend, linejoin, linemitre) {
    if (length(data$x) == 0) {
      return(NULL)
    }

    polylineGrob(
      data$x, data$y, data$id,
      default.units = x$units,
      gp = gpar(
        col = col, alpha = alpha, lty = lty, lwd = lwd, lex = lex,
        lineend = lineend, linejoin = linejoin, linemitre = linemitre
      )
    )
  }

  # merge current and grob-specific graphical parameters so we can redistribute among isolevels
  gp <- modifyList(x$gp_cur, x$gp_combined)
  n <- length(x$lines)

  lines_grobs <- mapply(
    make_lines_grobs,
    x$lines,
    rep_len(gp$col, n),
    rep_len(gp$alpha, n),
    rep_len(gp$lty, n),
    rep_len(gp$lwd, n),
    rep_len(gp$lex, n),
    rep_len(gp$lineend, n),
    rep_len(gp$linejoin, n),
    rep_len(gp$linemitre, n),
    SIMPLIFY = FALSE
  )

  children <- do.call(gList, lines_grobs)
  setChildren(x, children)
}

test_that("basic functions", {
  m <- matrix(c(0, 0, 0, 0, 0,
                0, 0, 1, 0, 0,
                0, 1, 2, 1, 0,
                0, 0, 1, 0, 0,
                0, 0, 0, 0, 0), 5, 5, byrow = TRUE)

  l <- isolines((1:5)/6, (5:1)/6, m, c(.5, 1.5))

  # incorrect number of labels
  expect_error(
    isolines_grob(l, labels = c("a", "b", "c")),
    "Number of labels must match the number of breaks."
  )

  # incorrect margin specification
  expect_error(
    isolines_grob(l, margin = 1:4),
    "must be a unit object of length four"
  )
  expect_error(
    isolines_grob(l, margin = grid::unit(1:3, "pt")),
    "must be a unit object of length four"
  )

  # default settings create two labels per line for this dataset
  g <- isolines_grob(l, label_placer = label_placer_minmax(n = 0))
  expect_equal(g$labels_data$break_id, c("0.5", "0.5", "1.5", "1.5"))
  expect_equal(g$labels_data$label, c("0.5", "0.5", "1.5", "1.5"))
  expect_equal(g$labels_data$x, rep(0.5, 4))
  expect_equal(g$labels_data$theta, rep(0, 4))
  expect_true(all(abs(g$labels_data$y - c(0.75, 0.25, 0.5833333, 0.4166667)) < 1e-7))

  # extra breaks are ignored
  g <- isolines_grob(l, breaks = c("0.5", "1.5", "2.5"), labels = c("a", "b", "c"))
  expect_equal(g$labels_data$break_id, c("0.5", "0.5", "1.5", "1.5"))
  expect_equal(g$labels_data$label, c("a", "a", "b", "b"))
})

#' Clip lines so they don't run into a set of boxes.
#'
#' Clip lines so they don't run into a set of boxes. Useful for labeling isolines,
#' as it allows removal of line segments that would run into any text labels.
#' @param x Numeric vector of x coordinates
#' @param y Numeric vector of y coordinates
#' @param id Integer vector of id numbers indicating which lines are connected
#' @param clip_boxes Data frame specifying the locations of boxes to clip to.
#'   Should have five columns, named `x`, `y`, `width`, `height`, `theta`, which
#'   specify the x and y positions of each box midpoint, as well as the box width,
#'   box height, and box angle in radians. Each box is specified by one data
#'   row.
#' @param asp Aspect ratio (width/height) of the target canvas. This is used to convert
#'   widths to heights and vice versa for rotated boxes
#' @export
clip_lines <- function(x, y, id, clip_boxes, asp = 1) {
  out = list(x = x, y = y, id = id)
  for (i in 1:nrow(clip_boxes)) {
    box <- clip_boxes[i, ]
    out <- clip_lines_impl(
      out$x, out$y, out$id,
      box$x, box$y, box$width, box$height, box$theta,
      asp
    )
  }

  out
}

clip_lines_impl <- function(x, y, id, p_mid_x, p_mid_y, width, height, theta, asp = 1) {
  .Call(
    `clip_lines_impl_c`, as.numeric(x), as.numeric(y), as.integer(id),
    as.numeric(p_mid_x), as.numeric(p_mid_y), as.numeric(width),
    as.numeric(height), as.numeric(theta), as.numeric(asp), PACKAGE = "isoband"
  )
}

test_that("basic clipping", {
  x <- c(0, 0, 1, 1, 0)
  y <- c(0, 1, 1, 0, 0)
  id <- rep(1L, 5)

  # clip box entirely inside line
  out <- clip_lines_impl(x, y, id, .5, .5, .8, .8, 0)
  expect_identical(out$x, x)
  expect_identical(out$y, y)
  expect_identical(out$id, id)

  # clip box entirely outside line
  out <- clip_lines_impl(x, y, id, .5, .5, 1.2, 1.2, 0)
  expect_identical(out$x, numeric(0))
  expect_identical(out$y, numeric(0))
  expect_identical(out$id, integer(0))

  # clip top right corner
  out <- clip_lines_impl(x, y, id, 1, 1, 1, 1, 0)
  expect_equal(out$x, c(0.0, 0.0, 0.5, 1.0, 1.0, 0.0))
  expect_equal(out$y, c(0.0, 1.0, 1.0, 0.5, 0.0, 0.0))
  expect_identical(out$id, c(rep(1L, 3), rep(2L, 3)))

  # clip bottom left corner
  out <- clip_lines_impl(x, y, id, 0, 0, 1, 1, 0)
  expect_equal(out$x, c(0.0, 0.0, 1.0, 1.0, 0.5))
  expect_equal(out$y, c(0.5, 1.0, 1.0, 0.0, 0.0))
  expect_identical(out$id, rep(1L, 5))

  # clip right half
  out <- clip_lines_impl(x, y, id, 1, .5, 1, 2, 0)
  expect_equal(out$x, c(0.0, 0.0, 0.5, 0.5, 0.0))
  expect_equal(out$y, c(0, 1, 1, 0, 0))
  expect_identical(out$id, c(rep(1L, 3), 2L, 2L))

  # clip left half
  out <- clip_lines_impl(x, y, id, 0, .5, 1, 2, 0)
  expect_equal(out$x, c(0.5, 1.0, 1.0, 0.5))
  expect_equal(out$y, c(1, 1, 0, 0))
  expect_identical(out$id, rep(1L, 4))

  # clip in middle
  out <- clip_lines_impl(x, y, id, .5, .5, 2, .5, 0)
  expect_equal(out$x, c(0, 0, 0, 0, 1, 1, 1, 1, 0))
  expect_equal(out$y, c(0.00, 0.25, 0.75, 1.00, 1.00, 0.75, 0.25, 0.00, 0.00))
  expect_identical(out$id, c(rep(1L, 2), rep(2L, 4), rep(3L, 3)))

  out <- clip_lines_impl(x, y, id, .5, .5, .5, 2, 0)
  expect_equal(out$x, c(0.00, 0.00, 0.25, 0.75, 1.00, 1.00, 0.75, 0.25, 0.00))
  expect_equal(out$y, c(0, 1, 1, 1, 1, 0, 0, 0, 0))
  expect_identical(out$id, c(rep(1L, 3), rep(2L, 4), rep(3L, 2)))
})

test_that("clip multiple line segments", {
  x <- c(0, 0, 1, 1, 2, 2, 3, 3)
  y <- c(0, 1, 1, 0, 2, 3, 3, 2)
  id <- c(rep(1L, 4), rep(2L, 4))

  out <- clip_lines_impl(x, y, id, 0, 0, 1, 3, 0)
  expect_equal(out$x, c(0.5, 1.0, 1.0, 2.0, 2.0, 3.0, 3.0))
  expect_equal(out$y, c(1, 1, 0, 2, 3, 3, 2))
  expect_identical(out$id, c(rep(1L, 3), rep(2L, 4)))

  out <- clip_lines_impl(x, y, id, 1.5, 1.5, 2, 4, 0)
  expect_equal(out$x, c(0.0, 0.0, 0.5, 2.5, 3.0, 3.0))
  expect_equal(out$y, c(0, 1, 1, 3, 3, 2))
  expect_identical(out$id, c(rep(1L, 3), rep(2L, 3)))
})

test_that("rotated clip box", {
  x <- c(0, 0, 1, 1, 0)
  y <- c(0, 1, 1, 0, 0)
  id <- rep(1L, 5)

  out <- clip_lines_impl(x, y, id, .5, .5, 5, sin(2*pi*45/360), 2*pi*45/360)
  expect_equal(out$x, c(0.0, 0.0, 0.5, 1.0, 1.0, 0.5))
  expect_equal(out$y, c(0.5, 1, 1, 0.5, 0, 0))
  expect_identical(out$id, c(rep(1L, 3), rep(2L, 3)))

  out <- clip_lines_impl(x, y, id, .5, .5, 5, sin(2*pi*45/360), -2*pi*45/360)
  expect_equal(out$x, c(0.0, 0.0, 0.5, 1.0, 1.0, 0.5, 0.0))
  expect_equal(out$y, c(0.0, 0.5, 1, 1, 0.5, 0, 0))
  expect_identical(out$id, c(rep(1L, 2), rep(2L, 3), rep(3L, 2)))

  out <- clip_lines_impl(x, y, id, .5, .5, sin(2*pi*45/360), 5, 2*pi*45/360)
  expect_equal(out$x, c(0.0, 0.0, 0.5, 1.0, 1.0, 0.5, 0.0))
  expect_equal(out$y, c(0.0, 0.5, 1, 1, 0.5, 0, 0))
  expect_identical(out$id, c(rep(1L, 2), rep(2L, 3), rep(3L, 2)))
})

test_that("singletons are carried over or clipped", {
  x <- c(0, 0, 1, 1, 2, 2, 3, 3)
  y <- c(0, 1, 1, 0, 2, 3, 3, 2)
  id <- c(1L, rep(2L, 3), rep(3L, 4))

  out <- clip_lines_impl(x, y, id, 10, 10, 1, 1, 0)
  expect_identical(out$x, x)
  expect_identical(out$y, y)
  expect_identical(out$id, id)

  out <- clip_lines_impl(x, y, id, 0, 0, .1, .1, 0)
  expect_identical(out$x, x[2:8])
  expect_identical(out$y, y[2:8])
  expect_identical(out$id, id[2:8] - 1L)

  x <- c(0, 0, 1, 1, 2, 2, 3, 3)
  y <- c(0, 1, 1, 0, 2, 3, 3, 2)
  id <- c(rep(1L, 3), 2L, rep(3L, 4))

  out <- clip_lines_impl(x, y, id, 10, 10, 1, 1, 0)
  expect_equal(out$x, x)
  expect_equal(out$y, y)
  expect_identical(out$id, id)

  out <- clip_lines_impl(x, y, id, 1, 0, .1, .1, 0)
  expect_identical(out$x, x[c(1:3, 5:8)])
  expect_identical(out$y, y[c(1:3, 5:8)])
  expect_identical(out$id, c(rep(1L, 3), rep(2L, 4)))

  x <- c(0, 0, 1, 1, 2, 2, 3, 3)
  y <- c(0, 1, 1, 0, 2, 3, 3, 2)
  id <- c(rep(1L, 3), rep(2L, 4), 3L)

  out <- clip_lines_impl(x, y, id, 10, 10, 1, 1, 0)
  expect_equal(out$x, x)
  expect_equal(out$y, y)
  expect_identical(out$id, id)

  out <- clip_lines_impl(x, y, id, 3, 2, .1, .1, 0)
  expect_identical(out$x, x[1:7])
  expect_identical(out$y, y[1:7])
  expect_identical(out$id, id[1:7])
})

test_that("empty or incorrect input", {
  out <- clip_lines_impl(numeric(0), numeric(0), integer(0), 3, 2, .1, .1, 0)
  expect_identical(out$x, numeric(0))
  expect_identical(out$y, numeric(0))
  expect_identical(out$id, integer(0))

  expect_error(
    clip_lines_impl(numeric(0), numeric(1), integer(0), 3, 2, .1, .1, 0),
    "must match"
  )
  expect_error(
    clip_lines_impl(numeric(0), numeric(0), integer(1), 3, 2, .1, .1, 0),
    "must match"
  )
  expect_error(
    clip_lines_impl(numeric(1), numeric(0), integer(0), 3, 2, .1, .1, 0),
    "must match"
  )
})

# convenience operator, don't want to import rlang just for this
"%||%" <- function(x, y) {
  if (is.null(x)) {
    y
  } else {
    x
  }
}

# evaluates all arguments
# (simpler than forcing each argument individually)
force_all <- function(...) list(...)

rethrow_interrupt <- function() {
  interrupt <- structure(list(), class = c("interrupt", "condition"))
  signalCondition(interrupt)
  invokeRestart("abort")
}

#' Efficient calculation of isolines and isobands from elevation grid
#'
#' @param x Numeric vector specifying the x locations of the grid points.
#' @param y Numeric vector specifying the y locations of the grid points.
#' @param z Numeric matrix specifying the elevation values for each grid point.
#' @param levels_low,levels_high Numeric vectors of minimum/maximum z values
#'   for which isobands should be generated. Any z values that are exactly
#'   equal to a value in `levels_low` are considered part of the corresponding
#'   isoband, but any z values that are exactly equal to a value in `levels_high`
#'   are not considered part of the corresponding isoband. In other words, the
#'   intervals specifying isobands are closed at their lower boundary and open
#'   at their upper boundary.
#' @seealso
#' [`plot_iso`]
#' @examples
#' library(grid)
#'
#' #' # one simple connected shape
#' m <- matrix(c(0, 0, 0, 0, 0, 0,
#'               0, 0, 0, 1, 1, 0,
#'               0, 0, 1, 1, 1, 0,
#'               0, 1, 1, 0, 0, 0,
#'               0, 0, 0, 1, 0, 0,
#'               0, 0, 0, 0, 0, 0), 6, 6, byrow = TRUE)
#'
#' df_bands <- isobands((1:ncol(m))/(ncol(m)+1), (nrow(m):1)/(nrow(m)+1), m, 0.5, 1.5)[[1]]
#' df_lines <- isolines((1:ncol(m))/(ncol(m)+1), (nrow(m):1)/(nrow(m)+1), m, 0.5)[[1]]
#' g <- expand.grid(x = (1:ncol(m))/(ncol(m)+1), y = (nrow(m):1)/(nrow(m)+1))
#' grid.newpage()
#' grid.points(g$x, g$y, default.units = "npc", pch = 19, size = unit(0.5, "char"))
#' grid.path(df_bands$x, df_bands$y, df_bands$id, gp = gpar(fill = "cornsilk", col = NA))
#' grid.polyline(df_lines$x, df_lines$y, df_lines$id)
#'
#' # a similar plot can be generated with the plot_iso() function,
#' # which is useful for exploring how the algorithm works
#' plot_iso(m, 0.5, 1.5)
#'
#' # NAs are ignored
#' m <- matrix(c(NA, NA, NA, 0, 0, 0,
#'               NA, NA, NA, 1, 1, 0,
#'                0,  0,  1, 1, 1, 0,
#'                0,  1,  1, 0, 0, 0,
#'                0,  0,  0, 1, 0, 0,
#'                0,  0,  0, 0, 0, 0), 6, 6, byrow = TRUE)
#' plot_iso(m, 0.5, 1.5)
#'
#' # two separate shapes
#' m <- matrix(c(0, 0, 1, 1,
#'               0, 1, 1, 1,
#'               1, 1, 0, 0,
#'               0, 0, 0.8, 0), 4, 4, byrow = TRUE)
#' plot_iso(m, 0.5, 1.5)
#'
#' # shape with hole
#' m <- matrix(c(0, 0, 0, 0, 0, 0,
#'               0, 1, 1, 1, 1, 0,
#'               0, 1, 2, 2, 1, 0,
#'               0, 1, 2, 2, 1, 0,
#'               0, 1, 1, 1, 1, 0,
#'               0, 0, 0, 0, 0, 0), 6, 6, byrow = TRUE)
#' plot_iso(m, 0.5, 1.5)
#' @export
isobands <- function(x, y, z, levels_low, levels_high) {
  nlow <- length(levels_low)
  nhigh <- length(levels_high)
  nmax <- max(nlow, nhigh)

  if ((nlow != nmax && nlow != 1) || (nhigh != nmax && nhigh != 1)) {
    stop("Vectors specifying isoband levels must be of equal length or of length 1", call. = FALSE)
  }
  levels_low <- rep_len(levels_low, nmax)
  levels_high <- rep_len(levels_high, nmax)

  # swap high and low levels when they're given in the wrong order
  idx <- levels_high < levels_low
  if (any(idx)) {
    levels_tmp <- levels_high
    levels_high[idx] <- levels_low[idx]
    levels_low[idx] <- levels_tmp[idx]
  }

  out <- isobands_impl(x, y, z, levels_low, levels_high)
  structure(
    stats::setNames(out, paste0(levels_low, ":", levels_high)),
    class = c("isobands", "iso")
  )
}

#' @rdname isobands
#' @param levels Numeric vector of z values for which isolines should be generated.
#' @export
isolines <- function(x, y, z, levels) {
  out <- isolines_impl(x, y, z, levels)
  structure(
    stats::setNames(out, levels),
    class = c("isolines", "iso")
  )
}

isobands_impl <- function(x, y, z, value_low, value_high) {
  mode(z) <- "numeric"
  .Call(
    `isobands_impl_c`, as.numeric(x), as.numeric(y), z,
    as.numeric(value_low), as.numeric(value_high), PACKAGE = "isoband"
  )
}

isolines_impl <- function(x, y, z, value) {
  mode(z) <- "numeric"
  .Call(
    `isolines_impl_c`, as.numeric(x), as.numeric(y), z,
    as.numeric(value), PACKAGE = "isoband"
  )
}

#' Render isobands
#'
#' This function generates a grid grob that represents isobands.
#'
#' @param bands Isobands, as produced by the [`isobands()`] function.
#' @param gp Grid graphical parameters. Parameters are recycled among
#'   the total number of bands drawn.
#' @param units A character string specifying the units in which to
#'   interpret the isobands coordinates. Defaults to `"npc"`.
#' @seealso
#' See [`isolines_grob()`] for drawing of isolines.
#' @examples
#' library(grid)
#'
#' viridis_pal <- colorRampPalette(
#'   c("#440154", "#414487", "#2A788E", "#22A884", "#7AD151", "#FDE725"),
#'   space = "Lab"
#' )
#'
#' x <- (1:ncol(volcano))/(ncol(volcano)+1)
#' y <- (nrow(volcano):1)/(nrow(volcano)+1)
#' bands <- isobands(x, y, volcano, 5*(18:38), 5*(19:39))
#'
#' b <- isobands_grob(
#'   bands,
#'   gp = gpar(col = "black", fill = viridis_pal(21), alpha = 0.5)
#' )
#'
#' grid.newpage()
#' grid.draw(b)
#' @export
isobands_grob <- function(bands, gp = gpar(), units = "npc") {
  gTree(
    bands = bands,
    gp_user = gp,
    units = units,
    cl = "isobands_grob"
  )
}

#' @export
makeContent.isobands_grob <- function(x) {
  make_bands_grobs <- function(data, col, fill, alpha, lty, lwd, lex, lineend, linejoin, linemitre) {
    if (length(data$x) == 0) {
      return(NULL)
    }
    pathGrob(
      data$x, data$y, data$id,
      default.units = x$units,
      gp = gpar(
        col = col, fill = fill, alpha = alpha, lty = lty, lwd = lwd, lex = lex,
        lineend = lineend, linejoin = linejoin, linemitre = linemitre
      )
    )
  }

  # merge current and grob-specific graphical parameters so we can redistribute among isolevels
  gp <- modifyList(get.gpar(), x$gp_user)
  n <- length(x$bands)

  bands_grobs <- mapply(
    make_bands_grobs,
    x$bands,
    rep_len(gp$col, n),
    rep_len(gp$fill, n),
    rep_len(gp$alpha, n),
    rep_len(gp$lty, n),
    rep_len(gp$lwd, n),
    rep_len(gp$lex, n),
    rep_len(gp$lineend, n),
    rep_len(gp$linejoin, n),
    rep_len(gp$linemitre, n),
    SIMPLIFY = FALSE
  )

  children <- do.call(gList, bands_grobs)
  setChildren(x, children)
}

#' Generic label placement function
#'
#' The simple label placer processes separate isolines independently and places
#' labels for each line using a placer function that does the actual placement work.
#' This label placer is not meant to be used by end users, but rather facilitates the
#' development of new label placers, such as [`label_placer_minmax()`].
#' @param lines Isolines object for which labels should be placed.
#' @param labels_data A data frame containing information about which labels should
#'   be placed.
#' @param placer_fun A function that takes an individual isoline plus its associated
#'   break id as input and returns a data frame specifying label positions. The data
#'   frame should have three columns called `x`, `y`, and `theta`. `x` and `y` specify
#'   the label position, and `theta` specifies the label angle in radians. The data
#'   frame can have multiple rows, which results in the same label being placed in
#'   multiple locations.
#' @keywords internal
#' @export
label_placer_simple <- function(lines, labels_data, placer_fun) {
  # Calculate the label position for one set of isolines (one level).
  #
  # The line data is specified as a list of x, y, id. The parameters `index`, `break_index`,
  # `break_id`, and `label` are provided simply so they can be added to the resulting data
  # frame holding label positions
  place_labels_impl <- function(line_data, index, break_index, break_id, label, placer_fun) {
    # return empty row if either missing line data or missing label
    if (length(line_data$x) == 0 || is.na(label)) {
      return(
        data.frame(
          index = integer(0),
          break_index = integer(0),
          break_id = character(0), label = character(0),
          x = numeric(0), y = numeric(0), theta = numeric(0),
          stringsAsFactors = FALSE
        )
      )
    }

    # calculate label position
    pos <- placer_fun(line_data, break_id)

    # return results
    if (nrow(pos) > 0) {
      data.frame(
        index = index,
        break_index = break_index,
        break_id = break_id,
        label = label,
        x = pos$x, y = pos$y, theta = pos$theta,
        stringsAsFactors = FALSE
      )
    } else {
      data.frame(
        index = integer(0),
        break_index = integer(0),
        break_id = character(0),
        label = character(0),
        x = numeric(0), y = numeric(0), theta = numeric(0),
        stringsAsFactors = FALSE
      )
    }
  }

  rows <- mapply(
    place_labels_impl,
    lines[labels_data$index],
    labels_data$index, # index of labeled lines in original list of lines, for matching of graphical parameters
    labels_data$break_index, # index into original list of breaks, for matching graphical parameters
    labels_data$break_id,
    labels_data$label,
    MoreArgs = list(placer_fun = placer_fun),
    SIMPLIFY = FALSE
  )
  Reduce(rbind, rows)
}



#' Set up a label placement strategy
#'
#' These functions set up various label placement strategies.
#'
#' `label_placer_minmax()` places labels at the horizontal or vertical minima or maxima of
#' the respective isolines.
#'
#' `label_placer_none()` places no labels at all.
#'
#' `label_placer_manual()` places labels at manually defined locations.
#' 
#' `label_placer_middle()` places labels at the middle of each isoline.
#'
#' @param placement String consisting of any combination of the letters
#'   "t", "r", "b", "l" indicating the placement of labels at the top,
#'   to the right, at the bottom, to the left of the isoline.
#' @param rot_adjuster Function that standardizes the rotation angles of the labels.
#'   See e.g. [`angle_halfcircle_bottom()`].
#' @param n Size of the point neighborhood over which the rotation angle should be
#'   calculated.
#' @rdname label_placer
#' @export
label_placer_minmax <- function(placement = "tb", rot_adjuster = angle_halfcircle_bottom(), n = 2) {
  force_all(placement, rot_adjuster, n)

  placer_fun <- function(line_data, ...) {
    # find location for labels
    idx <- stats::na.omit(
      c(
        which( # placement "top"
          isTRUE(grepl("t", placement, fixed = TRUE)) & line_data$y == max(line_data$y)
        )[1],
        which( # placement "bottom"
          isTRUE(grepl("b", placement, fixed = TRUE)) & line_data$y == min(line_data$y)
        )[1],
        which( # placement "left"
          isTRUE(grepl("l", placement, fixed = TRUE)) & line_data$x == min(line_data$x)
        )[1],
        which( # placement "right"
          isTRUE(grepl("r", placement, fixed = TRUE)) & line_data$x == max(line_data$x)
        )[1]
      )
    )

    out <- data.frame(x = numeric(0), y = numeric(0), theta = numeric(0))

    for (i in seq_along(idx)) {
      out[i, ] <- minmax_impl(line_data, idx[i], n)
    }

    # standardize rotation angles for text labels
    out$theta <- rot_adjuster(out$theta)

    out
  }

  # final placer function
  function(lines, labels_data) {
    label_placer_simple(lines, labels_data, placer_fun)
  }
}

# function that does all the work for the minmax label placer.
# requires a single index idx
minmax_impl <- function(data, idx, n) {
  # set of indices belonging to this label
  idx_set <- which(data$id == data$id[idx])
  idx_min <- min(idx_set)
  idx_max <- max(idx_set)

  # if the first and the last point are the same we wrap, otherwise we truncate
  if (data$x[idx_min] == data$x[idx_max] && data$y[idx_min] == data$y[idx_max]) {
    idx_range <- (idx_max - idx_min)
    i <- ((idx - n):(idx + n)-idx_min) %% idx_range + idx_min
  } else {
    i <- (max(idx - n, idx_min):min(idx + n, idx_max))
  }

  x <- data$x[i]
  y <- data$y[i]
  xave <- mean(x)
  yave <- mean(y)
  m <- cbind(x - xave, y - yave)
  v <- svd(m)$v
  list(x = xave, y = yave, theta = atan2(v[2], v[1]))
}

#' @rdname label_placer
#' @export
label_placer_none <- function() {
  function(...) {
    data.frame(
      index = integer(0),
      break_index = integer(0),
      break_id = character(0), label = character(0),
      x = numeric(0), y = numeric(0), theta = numeric(0),
      stringsAsFactors = FALSE
    )
  }
}

#' @param breaks Character vector specifying the isolines to be labeled,
#'   as in [`isolines_grob()`].
#' @param x,y,theta Numeric vectors specifying the x and y positions and
#'   angles (in radians) for each label corresponding to each break.
#' @rdname label_placer
#' @export
label_placer_manual <- function(breaks, x, y, theta) {
  # recycle all inputs to the same length
  # also has the side effect of forcing them
  n <- max(length(breaks), length(x), length(y), length(theta))
  breaks <- rep_len(breaks, n)
  x <- rep_len(x, n)
  y <- rep_len(y, n)
  theta <- rep_len(theta, n)

  placer_fun <- function(line_data, break_id) {
    idx <- (breaks == break_id)
    data.frame(x = x[idx], y = y[idx], theta = theta[idx])
  }

  # final placer function
  function(lines, labels_data) {
    label_placer_simple(lines, labels_data, placer_fun)
  }
}


#' @rdname label_placer
#' @export
label_placer_middle <- function(rot_adjuster = angle_halfcircle_bottom()) {
  placer_fun <- function(line_data, ...) {
    out <- data.frame(x = numeric(0), y = numeric(0), theta = numeric(0))

    # It identifies each isoline subdivision. For an individual isoline the id column identifies the number of subdivisions.
    line_sections <- unique(line_data$id)

    # Then the label is printed at the middle of each isoline subdivision.
    for (i in 1:length(line_sections)) {
      x <- line_data$x[line_data$id == i]
      y <- line_data$y[line_data$id == i]

      middle_index <- as.integer(length(x) / 2)

      x_mid <- x[middle_index]
      y_mid <- y[middle_index]

      xtheta <- c(x[middle_index - 1], x[middle_index], x[middle_index + 1])
      ytheta <- c(y[middle_index - 1], y[middle_index], y[middle_index + 1])

      m <- cbind(xtheta - mean(xtheta), ytheta - mean(ytheta))
      v <- svd(m)$v

      out[i, ] <- list(x = x_mid, y = y_mid, theta = atan2(v[2], v[1]))
    }
    # standardize rotation angles for text labels
    out$theta <- rot_adjuster(out$theta)
    out
  }

  # final placer function
  function(lines, labels_data) {
    label_placer_simple(lines, labels_data, placer_fun)
  }
}

#' Standardize label angles
#'
#' Function factories that return functions to standardize rotation angles to specific angle ranges.
#'
#' `angle_halfcircle_bottom()` standardizes angles to (-pi/2, pi/2].
#'
#' `angle_halfcircle_right()` standardizes angles to (0, pi].
#'
#' `angle_fixed()` sets all angles to a fixed value (0 by default).
#'
#' `angle_identity()` does not modify any angles.
#' @param theta Fixed angle, in radians.
#' @export
angle_halfcircle_bottom <- function() {
  function(theta) {
    ifelse(
      theta <= -pi/2,
      theta + pi,
      ifelse(
        theta > pi/2,
        theta - pi,
        theta
      )
    )
  }
}

#' @rdname angle_halfcircle_bottom
#' @export
angle_halfcircle_right <- function() {
  function(theta) {
    ifelse(
      theta <= 0,
      theta + pi,
      ifelse(
        theta > pi,
        theta - pi,
        theta
      )
    )
  }
}

#' @rdname angle_halfcircle_bottom
#' @export
angle_fixed <- function(theta = 0) {
  force(theta)

  function(x) {
    rep_len(theta, length(x))
  }
}

#' @rdname angle_halfcircle_bottom
#' @export
angle_identity <- function() {
  function(x) {
    x
  }
}

test_that("minmax label placer", {
  lines <- list(
    "1" = list(x = c(.5, .75, .5, .25), y = c(.25, .5, .75, .5), id = rep(1, 4)),
    "2" = list(x = c(.5, 1, .5, 0), y = c(0, .5, 1, .5), id = rep(1, 4))
  )

  labels_data <- data.frame(
    index = 1:2,
    break_index = c(3, 7),
    break_id = c("1", "2"),
    label = c("a", "b"),
    stringsAsFactors = FALSE
  )

  lp <- label_placer_minmax(n = 0)
  out <- lp(lines, labels_data)
  expect_equal(out$index, rep(1:2, each = 2))
  expect_equal(out$break_index, rep(c(3, 7), each = 2))
  expect_equal(out$break_id, rep(c("1", "2"), each = 2))
  expect_equal(out$label, rep(c("a", "b"), each = 2))
  expect_equal(out$x, rep(0.5, 4))
  expect_equal(out$y, c(0.75, 0.25, 1, 0))
  expect_equal(out$theta, rep(0, 4))

  lp <- label_placer_minmax(placement = "rl", rot_adjuster = angle_fixed(1.5), n = 0)
  out <- lp(lines, labels_data)
  expect_equal(out$x, c(0.25, 0.75, 0, 1))
  expect_equal(out$y, rep(0.5, 4))
  expect_equal(out$theta, rep(1.5, 4))

  lp <- label_placer_minmax(placement = NULL)
  out <- lp(lines, labels_data)
  expect_equal(nrow(out), 0)
})



test_that("angle adjustments", {
  theta <- c(-3, -2, -1, 0, 1, 2, 3)

  expect_equal(
    angle_halfcircle_bottom()(theta),
    c(-3 + pi, -2 + pi, -1, 0, 1, 2 - pi, 3 - pi)
  )

  expect_equal(
    angle_halfcircle_right()(theta + 2),
    c(-1 + pi, 0 + pi, 1, 2, 3, 4 - pi, 5 - pi)
  )

  expect_equal(
    angle_fixed()(theta),
    rep(0, length(theta))
  )

  expect_equal(
    angle_fixed(2)(theta),
    rep(2, length(theta))
  )

  expect_equal(
    angle_identity()(theta),
    theta
  )
})

test_that("none label placer", {
  lines <- list(
    "1" = list(x = c(.5, .75, .5, .25), y = c(.25, .5, .75, .5), id = rep(1, 4)),
    "2" = list(x = c(.5, 1, .5, 0), y = c(0, .5, 1, .5), id = rep(1, 4))
  )

  labels_data <- data.frame(
    index = 1:2,
    break_index = c(3, 7),
    break_id = c("1", "2"),
    label = c("a", "b"),
    stringsAsFactors = FALSE
  )

  lp <- label_placer_none()
  out <- lp(lines, labels_data)
  expect_equal(nrow(out), 0)
})


test_that("manual label placer", {
  lines <- list(
    "1" = list(x = c(.5, .75, .5, .25), y = c(.25, .5, .75, .5), id = rep(1, 4)),
    "2" = list(x = c(.5, 1, .5, 0), y = c(0, .5, 1, .5), id = rep(1, 4))
  )

  labels_data <- data.frame(
    index = 1:2,
    break_index = c(3, 7),
    break_id = c("1", "2"),
    label = c("a", "b"),
    stringsAsFactors = FALSE
  )

  lp <- label_placer_manual(
    breaks = c("1", "2", "3", "2"),
    x = 1:4,
    y = 4:1,
    theta = (1:4)-2
  )
  out <- lp(lines, labels_data)
  expect_equal(out$index, c(1, 2, 2))
  expect_equal(out$break_index, c(3, 7, 7))
  expect_equal(out$break_id, c("1", "2", "2"))
  expect_equal(out$label, c("a", "b", "b"))
  expect_equal(out$x, c(1, 2, 4))
  expect_equal(out$y, c(4, 3, 1))
  expect_equal(out$theta, c(-1, 0, 2))
})

# Two isolines id=1
test_that("middle label placer", {
  lines <- list(
    "1" = list(x = c(.5, .75, .5, .25), y = c(.25, .5, .75, .5), id = rep(1, 4)),
    "2" = list(x = c(.5, 1, .5, 0), y = c(0, .5, 1, .5), id = rep(1, 4))
  )

  labels_data <- data.frame(
    index = 1:2,
    break_index = c(3, 7),
    break_id = c("1", "2"),
    label = c("a", "b"),
    stringsAsFactors = FALSE
  )

  lp <- label_placer_middle()
  out <- lp(lines, labels_data)
  expect_equal(out$index, 1:2)
  expect_equal(out$break_index, c(3, 7))
  expect_equal(out$break_id, c("1", "2"))
  expect_equal(out$label, c("a", "b"))
  expect_equal(out$x, c(0.75, 1))
  expect_equal(out$y, c(0.5, 0.5))
  expect_equal(out$theta, c(pi / 2, pi / 2))
})

# One isoline with two id values (1 and 2)
test_that("middle label placer", {
  lines <- list(
    "1" = list(x = c(.5, .75, .5, .25, .5, 1, .5, 0), y = c(.25, .5, .75, .5, 0, .5, 1, .5), id = c(rep(1, 4), rep(2, 4)))
  )

  labels_data <- data.frame(
    index = 1,
    break_index = 3,
    break_id = "1",
    label = "a",
    stringsAsFactors = FALSE
  )

  lp <- label_placer_middle()
  out <- lp(lines, labels_data)
  expect_equal(out$index, c(1, 1))
  expect_equal(out$break_index, c(3, 3))
  expect_equal(out$break_id, c("1", "1"))
  expect_equal(out$label, c("a", "a"))
  expect_equal(out$x, c(0.75, 1))
  expect_equal(out$y, c(0.5, 0.5))
  expect_equal(out$theta, c(pi / 2, pi / 2))
})

#' Visualize a single isoband
#'
#' This function visualizes a single isoband calculated from a matrix. It is mainly useful
#' for debugging and visualizing the isobanding algorithm. See [`isobands()`] for more
#' examples.
#'
#' @param m input matrix
#' @param vlo lower cutoff for isobanding
#' @param vhi higher cutoff for isobanding
#' @param col_lo line color for lower cutoff
#' @param col_hi line color for higher cutoff
#' @param fill_lo fill color for points below the lower cutoff
#' @param fill_mid fill color for points between the two cutoffs
#' @param fill_hi fill color for points above the higher cutoff
#' @param fill_band fill color for the isoband
#' @param newpage boolean, indicating whether `grid.newpage()` should
#'   be called or not
#' @examples
#' m <- matrix(c(0, 0, 0, 0, 0, 0,
#'               0, 2, 2, 2, 2, 0,
#'               0, 2, 0, 0, 2, 0,
#'               0, 2, 0, 0, 2, 0,
#'               0, 2, 2, 2, 2, 0,
#'               0, 0, 0, 0, 0, 0), 6, 6, byrow = TRUE)
#'
#' plot_iso(m, 0.5, 1.5)
#' @export
plot_iso <- function(m, vlo, vhi, fill_lo = "gray95", fill_mid = "gray50", fill_hi = "black",
                     fill_band = "cornsilk", col_lo = "black", col_hi = "black", newpage = TRUE) {
  x <- 0.05 + 0.9*(0:(ncol(m)-1))/(ncol(m)-1)
  y <- 0.05 + 0.9*((nrow(m)-1):0)/(nrow(m)-1)
  df_bands <- isobands(x, y, m, vlo, vhi)[[1]]
  df_lines_lo <- isolines(x, y, m, vlo)[[1]]
  df_lines_hi <- isolines(x, y, m, vhi)[[1]]
  df_points <- expand.grid(y = y, x = x)
  pfill <- c(ifelse(m < vlo, fill_lo, ifelse(m < vhi, fill_mid, fill_hi)))
  pcol <- c(ifelse(m < vlo, "black", ifelse(m < vhi, fill_mid, "black")))
  if (isTRUE(newpage)) grid.newpage()
  if (length(df_bands$x) > 0)
    grid.path(df_bands$x, df_bands$y, df_bands$id, gp = gpar(fill = fill_band, col = NA))
  if (length(df_lines_lo$x) > 0)
    grid.polyline(df_lines_lo$x, df_lines_lo$y, df_lines_lo$id,
                  gp = gpar(col = col_lo))
  if (length(df_lines_hi$x) > 0)
    grid.polyline(df_lines_hi$x, df_lines_hi$y, df_lines_hi$id,
                  gp = gpar(col = col_hi))
  grid.points(df_points$x, df_points$y, default.units = "npc", pch = 21, size = unit(0.5, "char"),
              gp = gpar(fill = pfill, col = pcol))
}

---
title: "Generating isolines and isobands"
author: "Claus O. Wilke"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Generating isolines and isobands}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The isoband package implements fast algorithms for generating isolines (lines of equal elevation) and isobands (ranges of elevation delimited by two isolines) from a matrix of elevation data. For both cases, the package employs the marching squares algorithms as described on [Wikipedia.](https://en.wikipedia.org/wiki/Marching_squares) Marching squares algorithms break down the elevation matrix into blocks of 2x2 elevation values. For each block, they then determine the appropriate isolines/isobands from a lookup table of all possible arrangements of isolines or isobands within a 2x2 block. There are 16 distinct possibilities for isolines and 81 for isobands. The implementation in the isoband package goes beyond the algorithm described on Wikipedia in that it merges the isolines or isobands from separate blocks into extended line traces or polygons. The package is meant as a low-level package with minimal required dependencies. Therefore, many of the functions provided may not immediately be useful to endusers, but they will enable developers of other packages to integrate isolines and isobands into their feature set. 

The two main functions of the package are called `isolines()` and `isobands()`, and they have similar user interfaces and return values. Both take a vector `x` specifying the x values corresponding to the columns of the elevation matrix, a vector `y` specifying the y values corresponding to the rows of the elevation matrix, and an elevation matrix `z`. The two functions differ in that `isolines()` takes a single argument `levels` specifying the elevation levels for which isolines should be calculated, whereas `isobands()` takes two arguments, `levels_low` and `levels_high`, specifying the lower and upper bounds for each isoband. The return value in both cases is a list of lists. The outer list contains one list element for each specified isolevel. The inner lists hold line or polygon data in the form `x`, `y`, `id` as used by `grid::polylineGrob()` or `grid::pathGrob()`. The format has been chosen for easy drawing of the resulting values via these two grid functions.

```{r}
library(isoband)
library(grid)

m <- matrix(
  c(0, 0, 0, 0, 0,
    0, 1, 2, 1, 0,
    0, 1, 2, 0, 0,
    0, 1, 0, 1, 0,
    0, 0, 0, 0, 0),
  5, 5, byrow = TRUE
)

lines <- isolines(x = 1:ncol(m)/6, y = nrow(m):1/6, z = m, levels = 0.5)
lines
grid.newpage()
grid.draw(polylineGrob(lines[[1]]$x, lines[[1]]$y, lines[[1]]$id))

bands <- isobands(x = 1:ncol(m)/6, y = nrow(m):1/6, z = m, levels_low = 0.5, levels_high = 1.5)
bands
grid.newpage()
grid.draw(pathGrob(bands[[1]]$x, bands[[1]]$y, bands[[1]]$id, gp = gpar(fill = "cornsilk")))
```

A convenience function `plot_iso()` can be used to inspect a single isoband and corresponding isolines for an elevation matrix. This function is mostly meant for debugging and illustration purposes. It draws a grid of matrix points colored by whether each point is below, within, or above the isoband, as well as the isoband itself and the enclosing isolines.

```{r}
plot_iso(m, 0.5, 1.5)
```

The isoband package handles `NA` values in the matrix by simply ignoring the respective grid points.

```{r}
m <- matrix(
  c(NA, NA, NA, 0, 0, 0,
    NA, NA, NA, 1, 1, 0,
     0,  0,  1, 1, 1, 0,
     0,  1,  1, 0, 0, 0,
     0,  0,  0, 1, 0, 0,
     0,  0,  0, 0, 0, 0),
  6, 6, byrow = TRUE
)
plot_iso(m, 0.5, 1.5)
```

Isobands can contain holes, as shown above, and they can also consist of multiple disconnected pieces.

```{r}
m <- matrix(
  c(0, 0, 1, 1,
    0, 1, 1, 1,
    1, 1, 0, 0,
    0, 0, 0.8, 0),
  4, 4, byrow = TRUE
)
plot_iso(m, 0.5, 1.5)
```


# Performance

The code is written in C++ and performance is generally good. Isolining is about as fast as `grDevices::contourLines()`, isobanding is approximately 2.5 times slower.


```{r}
# contouring with contourLines() from grDevices
fn_contourLines <- function() {
  grDevices::contourLines(1:ncol(volcano), 1:nrow(volcano), volcano, levels = 10*(10:18))
}

# contouring with isolines()
fn_isolines <- function() {
  isolines(1:ncol(volcano), 1:nrow(volcano), volcano, 10*(10:18))
}

# contouring with isobands()
fn_isobands <- function() {
  isobands(1:ncol(volcano), 1:nrow(volcano), volcano, 10*(9:17), 10*(10:18))
}

microbenchmark::microbenchmark(fn_contourLines(), fn_isolines(), fn_isobands())
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/label-placer.R
\name{label_placer_minmax}
\alias{label_placer_minmax}
\alias{label_placer_none}
\alias{label_placer_manual}
\alias{label_placer_middle}
\title{Set up a label placement strategy}
\usage{
label_placer_minmax(
  placement = "tb",
  rot_adjuster = angle_halfcircle_bottom(),
  n = 2
)

label_placer_none()

label_placer_manual(breaks, x, y, theta)

label_placer_middle(rot_adjuster = angle_halfcircle_bottom())
}
\arguments{
\item{placement}{String consisting of any combination of the letters
"t", "r", "b", "l" indicating the placement of labels at the top,
to the right, at the bottom, to the left of the isoline.}

\item{rot_adjuster}{Function that standardizes the rotation angles of the labels.
See e.g. \code{\link[=angle_halfcircle_bottom]{angle_halfcircle_bottom()}}.}

\item{n}{Size of the point neighborhood over which the rotation angle should be
calculated.}

\item{breaks}{Character vector specifying the isolines to be labeled,
as in \code{\link[=isolines_grob]{isolines_grob()}}.}

\item{x, y, theta}{Numeric vectors specifying the x and y positions and
angles (in radians) for each label corresponding to each break.}
}
\description{
These functions set up various label placement strategies.
}
\details{
\code{label_placer_minmax()} places labels at the horizontal or vertical minima or maxima of
the respective isolines.

\code{label_placer_none()} places no labels at all.

\code{label_placer_manual()} places labels at manually defined locations.

\code{label_placer_middle()} places labels at the middle of each isoline.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/label-placer.R
\name{label_placer_simple}
\alias{label_placer_simple}
\title{Generic label placement function}
\usage{
label_placer_simple(lines, labels_data, placer_fun)
}
\arguments{
\item{lines}{Isolines object for which labels should be placed.}

\item{labels_data}{A data frame containing information about which labels should
be placed.}

\item{placer_fun}{A function that takes an individual isoline plus its associated
break id as input and returns a data frame specifying label positions. The data
frame should have three columns called \code{x}, \code{y}, and \code{theta}. \code{x} and \code{y} specify
the label position, and \code{theta} specifies the label angle in radians. The data
frame can have multiple rows, which results in the same label being placed in
multiple locations.}
}
\description{
The simple label placer processes separate isolines independently and places
labels for each line using a placer function that does the actual placement work.
This label placer is not meant to be used by end users, but rather facilitates the
development of new label placers, such as \code{\link[=label_placer_minmax]{label_placer_minmax()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iso-to-sfg.R
\name{iso_to_sfg}
\alias{iso_to_sfg}
\title{Convert isolines or isobands to sfg object}
\usage{
iso_to_sfg(x)
}
\arguments{
\item{x}{The object to convert.}
}
\description{
Convert isolines or isobands to an sf geometry collection (\code{sfg}) object. Further downstream
processing needs to happen via the sf package.
}
\details{
The function \code{iso_to_sfg()} is a generic that takes an object created by either \code{\link[=isolines]{isolines()}}
or \code{\link[=isobands]{isobands()}} and turns it into a simple features (sf) geometry collection. Importantly,
the isobanding algorithm can produce polygons that do not represent valid simple features. This
happens usually when the lower limit of an isoband is exactly equal to some data values (see
examples for a demonstration). This can be worked around either by slightly shifting the data
or band limits (e.g., round all data values and then shift them by a value smaller than the
rounding error) or by fixing the geometries using the function \code{st_make_valid()}.
}
\examples{
if (requireNamespace("sf", quietly = TRUE)) {
library(sf)
library(ggplot2)

# Example 1: simple 5x5 matrix
m <- matrix(c(0, 2, 2, 2, 0,
              0, 1, 0, 1, 0,
              0, 1, 0, 0, 0,
              0, 1, 0, 1, 0,
              0, 0, 0, 0, 0), 5, 5, byrow = TRUE)

z <- isolines(1:ncol(m), nrow(m):1, m, c(0.5, 1.5))
lines <- iso_to_sfg(z)
x <- st_sf(level = names(lines), geometry = st_sfc(lines))
ggplot(x) + geom_sf(aes(color = level))

# Example 2: volcano dataset
m <- volcano
b <- isobands((1:ncol(m))/(ncol(m)+1), (nrow(m):1)/(nrow(m)+1), m,
              10*9:19, 10*10:20)
bands <- iso_to_sfg(b)
x <- st_sf(level = as.numeric(sub(":.*", "", names(bands))), geometry = st_sfc(bands))
ggplot(x) + geom_sf(aes(color = level, fill = level))

# Example 3: invalid simple features
m <- matrix(c(1.5, 1.5, 1.5, 1.5, 0.6,
              0.5, 1.5, 1.5,   0,   0,
                0,   1,   0,   1,   1,
                0,   1,   0, 0.7,   0,
              0.9, 1.3, 1.8, 1.4, 0.4), 5, 5, byrow = TRUE)

raw <- isobands(1:5, 5:1, m, levels_low = 0:1, levels_high = 1:2)
bands <- iso_to_sfg(raw)

iso <- st_sf(
  id = factor(1:length(bands)),
  geometry = st_sfc(bands)
)

# the geometries are not valid
st_is_valid(iso, reason = TRUE)
# this doesn't prevent us from plotting them
ggplot(iso, aes(fill = id)) + geom_sf()

# make all geometries valid, requires GEOS >= 3.8.0
if (sf_extSoftVersion()["GEOS"] >= "3.8.0") {
  iso2 <- st_make_valid(iso)
  st_is_valid(iso2, reason=TRUE)
  # the plot should be unchanged
  ggplot(iso2, aes(fill = id)) + geom_sf()
}

# alternatively, if we shift all data values by a tiny
# amount (here, 1e-10) so they don't coincide with the band
# limits, no invalid geometries are generated.
raw <- isobands(1:5, 5:1, m + 1e-10, levels_low = 0:1, levels_high = 1:2)
bands <- iso_to_sfg(raw)
iso <- st_sf(id = factor(1:length(bands)), geometry = st_sfc(bands))
st_is_valid(iso, reason = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isoband.R
\docType{package}
\name{isoband-package}
\alias{isoband}
\alias{isoband-package}
\title{isoband: Generate Isolines and Isobands from Regularly Spaced Elevation Grids}
\description{
A fast C++ implementation to generate contour lines (isolines) and
  contour polygons (isobands) from regularly spaced grids containing elevation data.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://wilkelab.org/isoband/}
  \item Report bugs at \url{https://github.com/wilkelab/isoband/issues}
}

}
\author{
\strong{Maintainer}: Claus O. Wilke \email{wilke@austin.utexas.edu} (\href{https://orcid.org/0000-0002-7470-9261}{ORCID})

Authors:
\itemize{
  \item Thomas Lin Pedersen \email{thomasp85@gmail.com} (\href{https://orcid.org/0000-0002-5147-4711}{ORCID})
}

Other contributors:
\itemize{
  \item testthat and Catch authors (testthat C++ testing code) [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clip-lines.R
\name{clip_lines}
\alias{clip_lines}
\title{Clip lines so they don't run into a set of boxes.}
\usage{
clip_lines(x, y, id, clip_boxes, asp = 1)
}
\arguments{
\item{x}{Numeric vector of x coordinates}

\item{y}{Numeric vector of y coordinates}

\item{id}{Integer vector of id numbers indicating which lines are connected}

\item{clip_boxes}{Data frame specifying the locations of boxes to clip to.
Should have five columns, named \code{x}, \code{y}, \code{width}, \code{height}, \code{theta}, which
specify the x and y positions of each box midpoint, as well as the box width,
box height, and box angle in radians. Each box is specified by one data
row.}

\item{asp}{Aspect ratio (width/height) of the target canvas. This is used to convert
widths to heights and vice versa for rotated boxes}
}
\description{
Clip lines so they don't run into a set of boxes. Useful for labeling isolines,
as it allows removal of line segments that would run into any text labels.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isobands.R
\name{isobands}
\alias{isobands}
\alias{isolines}
\title{Efficient calculation of isolines and isobands from elevation grid}
\usage{
isobands(x, y, z, levels_low, levels_high)

isolines(x, y, z, levels)
}
\arguments{
\item{x}{Numeric vector specifying the x locations of the grid points.}

\item{y}{Numeric vector specifying the y locations of the grid points.}

\item{z}{Numeric matrix specifying the elevation values for each grid point.}

\item{levels_low, levels_high}{Numeric vectors of minimum/maximum z values
for which isobands should be generated. Any z values that are exactly
equal to a value in \code{levels_low} are considered part of the corresponding
isoband, but any z values that are exactly equal to a value in \code{levels_high}
are not considered part of the corresponding isoband. In other words, the
intervals specifying isobands are closed at their lower boundary and open
at their upper boundary.}

\item{levels}{Numeric vector of z values for which isolines should be generated.}
}
\description{
Efficient calculation of isolines and isobands from elevation grid
}
\examples{
library(grid)

#' # one simple connected shape
m <- matrix(c(0, 0, 0, 0, 0, 0,
              0, 0, 0, 1, 1, 0,
              0, 0, 1, 1, 1, 0,
              0, 1, 1, 0, 0, 0,
              0, 0, 0, 1, 0, 0,
              0, 0, 0, 0, 0, 0), 6, 6, byrow = TRUE)

df_bands <- isobands((1:ncol(m))/(ncol(m)+1), (nrow(m):1)/(nrow(m)+1), m, 0.5, 1.5)[[1]]
df_lines <- isolines((1:ncol(m))/(ncol(m)+1), (nrow(m):1)/(nrow(m)+1), m, 0.5)[[1]]
g <- expand.grid(x = (1:ncol(m))/(ncol(m)+1), y = (nrow(m):1)/(nrow(m)+1))
grid.newpage()
grid.points(g$x, g$y, default.units = "npc", pch = 19, size = unit(0.5, "char"))
grid.path(df_bands$x, df_bands$y, df_bands$id, gp = gpar(fill = "cornsilk", col = NA))
grid.polyline(df_lines$x, df_lines$y, df_lines$id)

# a similar plot can be generated with the plot_iso() function,
# which is useful for exploring how the algorithm works
plot_iso(m, 0.5, 1.5)

# NAs are ignored
m <- matrix(c(NA, NA, NA, 0, 0, 0,
              NA, NA, NA, 1, 1, 0,
               0,  0,  1, 1, 1, 0,
               0,  1,  1, 0, 0, 0,
               0,  0,  0, 1, 0, 0,
               0,  0,  0, 0, 0, 0), 6, 6, byrow = TRUE)
plot_iso(m, 0.5, 1.5)

# two separate shapes
m <- matrix(c(0, 0, 1, 1,
              0, 1, 1, 1,
              1, 1, 0, 0,
              0, 0, 0.8, 0), 4, 4, byrow = TRUE)
plot_iso(m, 0.5, 1.5)

# shape with hole
m <- matrix(c(0, 0, 0, 0, 0, 0,
              0, 1, 1, 1, 1, 0,
              0, 1, 2, 2, 1, 0,
              0, 1, 2, 2, 1, 0,
              0, 1, 1, 1, 1, 0,
              0, 0, 0, 0, 0, 0), 6, 6, byrow = TRUE)
plot_iso(m, 0.5, 1.5)
}
\seealso{
\code{\link{plot_iso}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/label-placer.R
\name{angle_halfcircle_bottom}
\alias{angle_halfcircle_bottom}
\alias{angle_halfcircle_right}
\alias{angle_fixed}
\alias{angle_identity}
\title{Standardize label angles}
\usage{
angle_halfcircle_bottom()

angle_halfcircle_right()

angle_fixed(theta = 0)

angle_identity()
}
\arguments{
\item{theta}{Fixed angle, in radians.}
}
\description{
Function factories that return functions to standardize rotation angles to specific angle ranges.
}
\details{
\code{angle_halfcircle_bottom()} standardizes angles to (-pi/2, pi/2].

\code{angle_halfcircle_right()} standardizes angles to (0, pi].

\code{angle_fixed()} sets all angles to a fixed value (0 by default).

\code{angle_identity()} does not modify any angles.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-iso.R
\name{plot_iso}
\alias{plot_iso}
\title{Visualize a single isoband}
\usage{
plot_iso(
  m,
  vlo,
  vhi,
  fill_lo = "gray95",
  fill_mid = "gray50",
  fill_hi = "black",
  fill_band = "cornsilk",
  col_lo = "black",
  col_hi = "black",
  newpage = TRUE
)
}
\arguments{
\item{m}{input matrix}

\item{vlo}{lower cutoff for isobanding}

\item{vhi}{higher cutoff for isobanding}

\item{fill_lo}{fill color for points below the lower cutoff}

\item{fill_mid}{fill color for points between the two cutoffs}

\item{fill_hi}{fill color for points above the higher cutoff}

\item{fill_band}{fill color for the isoband}

\item{col_lo}{line color for lower cutoff}

\item{col_hi}{line color for higher cutoff}

\item{newpage}{boolean, indicating whether \code{grid.newpage()} should
be called or not}
}
\description{
This function visualizes a single isoband calculated from a matrix. It is mainly useful
for debugging and visualizing the isobanding algorithm. See \code{\link[=isobands]{isobands()}} for more
examples.
}
\examples{
m <- matrix(c(0, 0, 0, 0, 0, 0,
              0, 2, 2, 2, 2, 0,
              0, 2, 0, 0, 2, 0,
              0, 2, 0, 0, 2, 0,
              0, 2, 2, 2, 2, 0,
              0, 0, 0, 0, 0, 0), 6, 6, byrow = TRUE)

plot_iso(m, 0.5, 1.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isolines-grob.R
\name{isolines_grob}
\alias{isolines_grob}
\title{Render labeled isolines}
\usage{
isolines_grob(
  lines,
  gp = gpar(),
  breaks = NULL,
  labels = NULL,
  margin = unit(c(1, 1, 1, 1), "pt"),
  label_col = NULL,
  label_alpha = NULL,
  label_placer = label_placer_minmax(),
  units = "npc"
)
}
\arguments{
\item{lines}{Isolines, as produced by the \code{\link[=isolines]{isolines()}} function.}

\item{gp}{Grid graphical parameters. Parameters applying to lines
(such as \code{col}, \code{lwd}, \code{lty}, etc.) are recycled among the total
number of lines drawn. Parameters applying only to labels (such
as \code{fontfamily}, \code{fontsize}) are recycled among the specified
breaks only. The two parameters \code{col} and \code{alpha} are also applied
to labels, unless overridden (see \code{label_col} and \code{label_alpha}),
but are matched to the corresponding lines.}

\item{breaks}{Character vector specifying the isolines that should be
labeled. If \code{NULL}, labels all isolines.}

\item{labels}{Character vector specifying the labels for each break.
If \code{NULL}, uses the breaks as labels. The number of labels provided
must match the number of breaks provided.}

\item{margin}{Unit object of length 4 specifying the top, right, bottom,
and left margins around each text label. The same margins are applied
to all labels.}

\item{label_col}{Color applied to labels. Can be used to override the
color provided in \code{gp}, in case labels and lines should have different
colors.}

\item{label_alpha}{Alpha applied to labels. Can be used to override the
alpha value provided in \code{gp}, in case labels and lines should have
different alpha values.}

\item{label_placer}{Function that controls how labels are placed along
the isolines. Uses \code{\link[=label_placer_minmax]{label_placer_minmax()}} by default.}

\item{units}{A character string specifying the units in which to
interpret the isolines coordinates. Defaults to \code{"npc"}.}
}
\description{
This function generates a grid grob that represents labeled isolines.
}
\examples{
library(grid)

viridis_pal <- colorRampPalette(
  c("#440154", "#414487", "#2A788E", "#22A884", "#7AD151", "#FDE725"),
  space = "Lab"
)

x <- (1:ncol(volcano))/(ncol(volcano)+1)
y <- (nrow(volcano):1)/(nrow(volcano)+1)
lines <- isolines(x, y, volcano, 5*(19:38))
bands <- isobands(x, y, volcano, 5*(18:38), 5*(19:39))

b <- isobands_grob(
  bands,
  gp = gpar(col = NA, fill = viridis_pal(21), alpha = 0.4)
)
l <- isolines_grob(
  lines, breaks = 20*(5:10),
  gp = gpar(
    lwd = c(.3, 1, .3, .3)
  )
)

grid.newpage()
grid.draw(b)
grid.draw(l)
}
\seealso{
See \code{\link[=isobands_grob]{isobands_grob()}} for drawing of isobands. See \code{\link[=label_placer_minmax]{label_placer_minmax()}} for
label placement strategies.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isobands-grob.R
\name{isobands_grob}
\alias{isobands_grob}
\title{Render isobands}
\usage{
isobands_grob(bands, gp = gpar(), units = "npc")
}
\arguments{
\item{bands}{Isobands, as produced by the \code{\link[=isobands]{isobands()}} function.}

\item{gp}{Grid graphical parameters. Parameters are recycled among
the total number of bands drawn.}

\item{units}{A character string specifying the units in which to
interpret the isobands coordinates. Defaults to \code{"npc"}.}
}
\description{
This function generates a grid grob that represents isobands.
}
\examples{
library(grid)

viridis_pal <- colorRampPalette(
  c("#440154", "#414487", "#2A788E", "#22A884", "#7AD151", "#FDE725"),
  space = "Lab"
)

x <- (1:ncol(volcano))/(ncol(volcano)+1)
y <- (nrow(volcano):1)/(nrow(volcano)+1)
bands <- isobands(x, y, volcano, 5*(18:38), 5*(19:39))

b <- isobands_grob(
  bands,
  gp = gpar(col = "black", fill = viridis_pal(21), alpha = 0.5)
)

grid.newpage()
grid.draw(b)
}
\seealso{
See \code{\link[=isolines_grob]{isolines_grob()}} for drawing of isolines.
}
